// -*- C++ -*-
/*!
 * @file  GPSReceive.cpp
 * @brief ModuleDescription
 * @date $Date$
 *
 * $Id$
 */

#include "GPSReceive.h"

// Module specification
// <rtc-template block="module_spec">
static const char* gpsreceive_spec[] =
  {
    "implementation_id", "GPSReceive",
    "type_name",         "GPSReceive",
    "description",       "ModuleDescription",
    "version",           "1.0.0",
    "vendor",            "MEDIA DRIVE",
    "category",          "Sensor",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.GPS_PORT", "/dev/ttyUSB0",
    "conf.default.GPS_BAUDRATE", "4800",

    // Widget
    "conf.__widget__.GPS_PORT", "text",
    "conf.__widget__.GPS_BAUDRATE", "text",
    // Constraints

    "conf.__type__.GPS_PORT", "string",
    "conf.__type__.GPS_BAUDRATE", "int",

    ""
  };
// </rtc-template>

int fd;
#define BUFFSIZE 255

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
GPSReceive::GPSReceive(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_gps_outOut("gps_out", m_gps_out)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
GPSReceive::~GPSReceive()
{
}



RTC::ReturnCode_t GPSReceive::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  
  // Set OutPort buffer
  addOutPort("gps_out", m_gps_outOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("GPS_PORT", m_GPS_PORT, "/dev/ttyUSB0");
  bindParameter("GPS_BAUDRATE", m_GPS_BAUDRATE, "4800");
  // </rtc-template>
  
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t GPSReceive::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GPSReceive::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GPSReceive::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t GPSReceive::onActivated(RTC::UniqueId ec_id)
{
  fd = open(m_GPS_PORT.c_str(), O_RDWR | O_NOCTTY);
  if (fd < 0) {
      perror(m_GPS_PORT.c_str());
      return RTC::RTC_ERROR;
  }

  m_gps_out.tm.sec = 0;
  m_gps_out.tm.nsec = 0;


  struct termios oldtio, newtio;
  tcgetattr(fd, &oldtio);
  memset(&newtio, 0, sizeof(newtio));
  newtio.c_cflag = CS8 | CLOCAL | CREAD;

  newtio.c_cc[VTIME] = 0;
  newtio.c_lflag = ICANON;
  newtio.c_iflag = IGNPAR | ICRNL;

   // ボーレートの設定
   cfsetispeed(&newtio, m_GPS_BAUDRATE);
   cfsetospeed(&newtio, m_GPS_BAUDRATE);
   // デバイスに設定を行う
   tcsetattr(fd, TCSANOW, &newtio);

  return RTC::RTC_OK;
}


RTC::ReturnCode_t GPSReceive::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t GPSReceive::onExecute(RTC::UniqueId ec_id)
{
    char buf[BUFFSIZE];
    int count;
    int utc;
    int msec;
    char stat;
    char *p;
    char ns;
    char ew;
    double lat;
    double lon;
    double sp;
    double he;
    int day;
    int mon;
    int year;

    memset(&buf, 0, sizeof(buf));
    count = read(fd, &buf, BUFFSIZE);
    if (count < 0) {
      std::cout << "CHILD:Could not read from serial port" << std::endl;
      return RTC::RTC_ERROR;
     } else {
      if ((p=strtok(buf,","))==NULL);
      if (strcmp(p,"$GPRMC")!=0)
        return RTC::RTC_OK; /*センテンスの先頭は$GPRMCか？*/
      if ((p=strtok(NULL,","))==NULL); /*時刻*/
      sscanf(p,"%d.%d",&utc, &msec);
      if ((p=strtok(NULL,","))==NULL); /*stat:ステータス*/
      stat=p[0];
      if ((p=strtok(NULL,","))==NULL); /*lat:緯度*/
      sscanf(p,"%lf",&lat);
      if ((p=strtok(NULL,","))==NULL); /*ns:南北*/
      ns=p[0];
      if ((p=strtok(NULL,","))==NULL); /*lon:経度*/
      sscanf(p,"%lf",&lon);
      if ((p=strtok(NULL,","))==NULL); /*ew:東西*/
      ew=p[0];
      if ((p=strtok(NULL,","))==NULL); /*速度*/
      sscanf(p,"%lf",&sp);
      if ((p=strtok(NULL,","))==NULL); /*進行方向*/
      sscanf(p,"%lf",&he);
      if ((p=strtok(NULL,","))==NULL); /*日付DDMMYY*/
      sscanf(p,"%2d%2d%2d",&day,&mon,&year);

      //std::cout << "CHILD:RCVD CHAR " << buf << " " << count;

       time_t utc_epoch_time;
       //time_t jst_epoch_time;
       struct tm tm;
       tm.tm_year = year + 100; // year start at 1900
       tm.tm_mon = mon - 1;  // january as 0
       tm.tm_mday = day;
       tm.tm_hour = utc/10000;
       tm.tm_min =  (utc%10000)/100;
       tm.tm_sec = utc%100;

       utc_epoch_time = timegm(&tm); // UTC to epoch

       //tm.tm_hour += 9;
       //jst_epoch_time = mktime(&tm); // JST to epoch

       RTC::GPSTime gpstime;
       gpstime.sec = utc_epoch_time;
       gpstime.msec = msec;

       m_gps_out.timeFromGPS = gpstime;
       m_gps_out.latitude = floor(lat/100.0) + fmod(lat,100.0)/60.0; // ddd.dddd変換
       if (ns=='S')  m_gps_out.latitude *= -1;
       m_gps_out.longitude = floor(lon/100.0) + fmod(lon,100.0)/60.0; // ddd.dddd変換
       if (ew=='W')  m_gps_out.longitude *= -1;

       //m_gps_out.altitude
       //m_gps_out.horizontalError
       //m_gps_out.verticalError
       //m_gps_out.heading
       //m_gps_out.horizontalSpeed
       //m_gps_out.verticalSpeed
       //m_gps_out.numSatellites
       //m_gps_out.fixType

       /*時刻緯度経度に対応するトークン表示*/
       printf("20%02d年%02d月%02d日 %02d時%02d分%02d秒(UTC) -> TIMESTAMP = %ld\n", year, mon, day ,utc/10000, (utc%10000)/100, utc%100, utc_epoch_time);/*utc*/
       //printf("20%02d年%02d月%02d日 %02d時%02d分%02d秒(JST) -> TIMESTAMP = %ld\n", year, mon, day ,utc/10000+9, (utc%10000)/100, utc%100, jst_epoch_time);/*utc+9時間は日本標準時*/
       if(stat=='A' || stat=='B') printf(" o "); else if(stat=='V' || stat=='W') printf(" x ");
       if (ns=='N') printf("北緯"); else if (ns=='S') printf("南緯");
       printf("%.0lf度%.4lf分",floor(lat/100.0), fmod(lat,100.0));
       if (ew=='E') printf("東経"); else if (ns=='W') printf("西経");
       printf("%.0lf度%.4lf分",floor(lon/100.0), fmod(lon,100.0));
       printf("\n");

      if(stat=='A' || stat=='B')
      {
        m_gps_outOut.write();
      }
    }


  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t GPSReceive::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GPSReceive::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GPSReceive::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GPSReceive::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GPSReceive::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void GPSReceiveInit(RTC::Manager* manager)
  {
    coil::Properties profile(gpsreceive_spec);
    manager->registerFactory(profile,
                             RTC::Create<GPSReceive>,
                             RTC::Delete<GPSReceive>);
  }
  
};


